from ir_measures import measures
from .base import Measure, ParamInfo


class _Bpref(measures.Measure):
    """
    Binary Preference (Bpref).
    This measure examines the relative ranks of judged relevant and non-relevant documents. Non-judged documents are not considered. 

    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/sigir/BuckleyV04,
          author       = {Chris Buckley and
                          Ellen M. Voorhees},
          title        = {Retrieval evaluation with incomplete information},
          booktitle    = {{SIGIR} 2004: Proceedings of the 27th Annual International {ACM} {SIGIR}
                          Conference on Research and Development in Information Retrieval, Sheffield,
                          UK, July 25-29, 2004},
          pages        = {25--32},
          publisher    = {{ACM}},
          year         = {2004},
          url          = {https://doi.org/10.1145/1008992.1009000},
          doi          = {10.1145/1008992.1009000}
        }
    """
    __name__ = 'Bpref'
    NAME = __name__
    PRETTY_NAME = 'Binary Preference'
    SHORT_DESC = 'The relative ranks of judged relevant and non-relevant documents.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


Bpref = _Bpref()
BPref = Bpref
measures.register(Bpref, ['BPref'])



import unittest
import itertools
import ir_measures
from ir_measures import Metric
from .base import BaseMeasureTest


class TestPytrecEval(BaseMeasureTest):

    def test_NumRet(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.NumRet
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 5)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 4)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 9)

        measure = ir_measures.NumRet(rel=1)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 3)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 1)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 4)

        measure = ir_measures.NumRet(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 1)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 2)

        measure = ir_measures.NumRet(rel=3)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0)

    def test_NumQ(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.NumQ
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 1)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 2)

    def test_NumRel(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.NumRel
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 3)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 2)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 5)


    def test_Success(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 2
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.Success@2
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.5)

        measure = ir_measures.Success(rel=2)@2
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.5)

        measure = ir_measures.Success(rel=3)@2
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0)

    def test_SetP(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 2
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.SetP(rel=1)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, .6)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.425)

        measure = ir_measures.SetRelP(rel=1)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .5)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.75)

        measure = ir_measures.SetP(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, .4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.325)

        measure = ir_measures.SetRelP(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 1.0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 1.0)

        measure = ir_measures.SetP(rel=3)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0)

    def test_SetR(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 2
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.SetR(rel=1)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .5)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.75)

        measure = ir_measures.SetR(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 1.0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 1.0)

        measure = ir_measures.SetR(rel=3)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0)

    def test_SetAP(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 2
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.SetAP(rel=1)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.6)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.125)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.3625)

        measure = ir_measures.SetAP(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.25)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.325)

        measure = ir_measures.SetAP(rel=3)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0)

    def test_SetF(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 2
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.SetF(rel=1)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.75, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertAlmostEqual(result[1].value, .33333, places=4)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.5417, places=4)

        measure = ir_measures.SetF(rel=1, beta=0.5)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.6923, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.3)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.49615, places=4)

        measure = ir_measures.SetF(rel=1, beta=2.0)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.81818, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.375)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.59659, places=4)

        measure = ir_measures.SetF(rel=3)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0)

        # make sure the multiple invocations hapen correctly
        res = provider.calc_aggregate([ir_measures.SetF(rel=1), ir_measures.SetF(rel=1, beta=0.5), ir_measures.SetF(rel=1, beta=2.0), ir_measures.SetF(rel=3)], qrels, run)
        self.assertAlmostEqual(res[ir_measures.SetF(rel=1)], 0.5417, places=4)
        self.assertAlmostEqual(res[ir_measures.SetF(rel=1, beta=0.5)], 0.49615, places=4)
        self.assertAlmostEqual(res[ir_measures.SetF(rel=1, beta=2.0)], 0.59659, places=4)
        self.assertEqual(res[ir_measures.SetF(rel=3)], 0)


    def test_IPrec(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 1
0 0 D1 1
0 0 D2 2
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.IPrec@0.25
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.625)

        measure = ir_measures.IPrec@0.5
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.625)

        measure = ir_measures.IPrec@0.75
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.0)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0)
        self.assertEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.5)

        measure = ir_measures.IPrec(rel=2)@0.1
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, .6666666, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.4583, places=4)

        measure = ir_measures.IPrec(rel=2)@0.25
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, .6666666, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.4583, places=4)

        measure = ir_measures.IPrec(rel=2)@0.5
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, .6666666, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.4583, places=4)

        measure = ir_measures.IPrec(rel=2)@0.75
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, .6666666, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .25)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.4583, places=4)


    def test_infAP(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 1
0 0 D1 -1
0 0 D2 0
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D4 -1
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.AP
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.8333, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, .125)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.47916666666666663, places=4)

        measure = ir_measures.infAP
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.8333, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.1875)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.510416, places=4)

        provider = ir_measures.providers.PytrecEvalProvider()
        measure = ir_measures.AP(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.33333, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertAlmostEqual(result[1].value, 0.25, places=4)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.2917, places=4)

        measure = ir_measures.infAP(rel=2)
        result = list(provider.iter_calc([measure], qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertAlmostEqual(result[0].value, 0.33333, places=4)
        self.assertEqual(result[1].query_id, "1")
        self.assertAlmostEqual(result[1].value, 0.375, places=4)
        self.assertAlmostEqual(provider.calc_aggregate([measure], qrels, run)[measure], 0.3542, places=4)

    def test_nDCG(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 1
0 0 D1 -1
0 0 D2 0
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D4 -1
1 0 D5 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D4 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.pytrec_eval
        measure = ir_measures.nDCG
        self.assertMetrics(
                provider.iter_calc([measure], qrels, run),
                [Metric(query_id='0', measure=measure, value=0.76018),
                Metric(query_id='1', measure=measure, value=0.32739)])

        measure = ir_measures.nDCG@3
        self.assertMetrics(
                provider.iter_calc([measure], qrels, run),
                [Metric(query_id='0', measure=measure, value=0.76018),
                Metric(query_id='1', measure=measure, value=0.0)])

        measure = ir_measures.nDCG(gains={0:1,1:4})
        self.assertMetrics(
                provider.iter_calc([measure], qrels, run),
                [Metric(query_id='0', measure=measure, value=0.97177),
                Metric(query_id='1', measure=measure, value=0.14949)])

    def test_P(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 -1
0 0 D1 1
0 0 D3 2
0 0 D4 1
1 0 D0 1
1 0 D3 0
1 0 D4 -1
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D0 2 0.7 run
1 0 D3 3 0.3 run
1 0 D2 4 0.4 run
'''))
        provider = ir_measures.pytrec_eval
        measure = ir_measures.P@4
        self.assertMetrics(
                provider.iter_calc([measure], qrels, run),
                [Metric(query_id='0', measure=measure, value=0.5),
                Metric(query_id='1', measure=measure, value=0.25)])

        measure = ir_measures.P(judged_only=True)@4
        self.assertMetrics(
                provider.iter_calc([measure], qrels, run),
                [Metric(query_id='0', measure=measure, value=0.75),
                Metric(query_id='1', measure=measure, value=0.25)])

    def test_judged_only(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 -1
0 0 D1 1
0 0 D3 2
0 0 D4 1
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
'''))
        run_unudged_removed = list(ir_measures.read_trec_run('''
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
''')) # D0 is removed because trec_eval considers negative labels as unjudged (??). See <https://github.com/usnistgov/trec_eval/blob/d95ca64e14a47d763ae349fb65e6d8cde4141dbd/form_res_rels.c#L219>
        provider = ir_measures.pytrec_eval
        for measure in [ir_measures.P@4, ir_measures.RR, ir_measures.Rprec, ir_measures.AP, ir_measures.AP@4, ir_measures.nDCG, ir_measures.nDCG@4, ir_measures.R@2, ir_measures.SetP, ir_measures.SetAP, ir_measures.SetF, ir_measures.Success@2, ir_measures.IPrec@0.5]:
            with self.subTest(measure):
                self.assertMetrics(
                        [Metric(m.query_id, measure, m.value) for m in provider.iter_calc([measure(judged_only=True)], qrels, run)],
                        provider.iter_calc([measure], qrels, run_unudged_removed))
                self.assertMetrics(
                        [Metric(m.query_id, measure, m.value) for m in provider.iter_calc([measure(judged_only=True)], qrels, run)],
                        provider.iter_calc([measure], qrels, run),
                        not_equal=True)

if __name__ == '__main__':
    unittest.main()



import contextlib
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED


class JudgedProvider(providers.Provider):
    """
    python implementation of judgment rate

    Adapted from OpenNIR's implementation: https://github.com/Georgetown-IR-Lab/OpenNIR/blob/master/onir/metrics/judged.py
    """
    NAME = 'judged'
    SUPPORTED_MEASURES = [
        measures._Judged(cutoff=Any())
    ]

    def _evaluator(self, measures, qrels):
        cutoffs = []
        for measure in ir_measures.util.flatten_measures(measures):
            if measure.NAME == 'Judged':
                cutoffs.append((measure['cutoff'], measure))
            else:
                raise ValueError(f'unsupported measure {measure}')
        qrels = ir_measures.util.QrelsConverter(qrels).as_dict_of_dict()
        return JudgedEvaluator(measures, qrels, cutoffs)


class JudgedEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, cutoffs):
        super().__init__(measures, set(qrels.keys()))
        self.qrels = qrels
        self.cutoffs = cutoffs

    def _iter_calc(self, run):
        run = ir_measures.util.RunConverter(run).as_dict_of_dict()
        sorted_run = {q: list(sorted(run[q].items(), key=lambda x: (-x[1], x[0]))) for q in run}
        for qid in run:
            qid_qrels = self.qrels.get(qid)
            if qid_qrels:
                for cutoff, measure in self.cutoffs:
                    current_run = sorted_run.get(qid, [])
                    # When there is no cutoff, default to the
                    # size of the run.
                    if cutoff == NOT_PROVIDED:
                        cutoff = len(current_run)

                    cutoff_run = current_run[:cutoff]
                    judged_c = sum((did in qid_qrels) for did, _ in cutoff_run)

                    # The cutoff should be recalculated if it is
                    # less than the size of then run.
                    if len(cutoff_run) < cutoff:
                        cutoff = len(cutoff_run)

                    # A cutoff larger than the run size causes
                    # this calculation to be incorrect.
                    value = judged_c / cutoff

                    yield Metric(query_id=qid, measure=measure, value=value)


providers.register(JudgedProvider())



from ir_measures import measures
from .base import Measure, ParamInfo


class _P(measures.Measure):
    """
    Basic measure for that computes the percentage of documents in the top cutoff results
    that are labeled as relevant. cutoff is a required parameter, and can be provided as
    P@cutoff.


    .. code-block:: bibtex
        :caption: Citation

        @book{DBLP:books/bu/Rijsbergen79,
          author       = {C. J. van Rijsbergen},
          title        = {Information Retrieval},
          publisher    = {Butterworth},
          year         = {1979},
          isbn         = {0-408-70929-4}
        }
    """
    __name__ = 'P'
    NAME = __name__
    PRETTY_NAME = 'Precision at k'
    SHORT_DESC = 'The percentage of documents in the top k results that are relevant.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


P = _P()
Precision = P
measures.register(P, ['Precision'])



from ir_measures import measures
from .base import Measure, ParamInfo


class _SetP(measures.Measure):
    """
    The Set Precision (SetP); i.e., the number of relevant docs divided by the total number retrieved
    """
    __name__ = 'SetP'
    NAME = __name__
    PRETTY_NAME = 'Set Precision'
    SHORT_DESC = 'The precision among all returned documents.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'relative': measures.ParamInfo(dtype=bool, default=False, desc='calculate the measure using the maximum possible SetP for the provided result size'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }

SetP = _SetP()
SetRelP = _SetP(relative=True)
measures.register(SetP)
measures.register(SetRelP, name='SetRelP')


class _SetR(measures.Measure):
    """
    The Set Recall (SetR); i.e., the number of relevant docs divided by the total number of relevant documents
    """
    __name__ = 'SetR'
    PRETTY_NAME = 'Set Recall'
    SHORT_DESC = 'The recall among all returned documents.'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }

SetR = _SetR()
measures.register(SetR)

class _SetF(measures.Measure):
    """
    The Set F measure (SetF); i.e., the harmonic mean of SetP and SetR
    """
    __name__ = 'SetF'
    PRETTY_NAME = 'Set F1'
    SHORT_DESC = 'The F1 score among all returned documents'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'beta': measures.ParamInfo(dtype=float, default=1., desc='relative importance of R to P in the harmonic mean'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }

SetF = _SetF()
measures.register(SetF)

class _SetAP(measures.Measure):
    """
    The unranked Set AP (SetAP); i.e., SetP * SetR
    """
    __name__ = 'SetAP'
    PRETTY_NAME = 'Set Average Precision'
    SHORT_DESC = 'The Average Precision among all returned documents'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }

SetAP = _SetAP()
measures.register(SetAP)



from ir_measures import measures
from .base import Measure, ParamInfo


class _Rprec(measures.Measure):
    """
    The precision at R, where R is the number of relevant documents for a given query. Has the cute property that
    it is also the recall at R.

    .. code-block:: bibtex
        :caption: Citation

        @book{841681,
          author = {C Buckley and Ellen Voorhees},
          title = {Retrieval System Evaluation},
          year = {2005},
          month = {2005-09-26 00:09:00},
          publisher = {TREC   Chapter to be published: TREC: Experiment and Evaluation in Information Retrieval, 2005, }
        }
    """
    __name__ = 'Rprec'
    NAME = __name__
    PRETTY_NAME = 'Precsion at R'
    SHORT_DESC = 'Precsion at R, where R is the number of relevant documents for a given query.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


Rprec = _Rprec()
RPrec = Rprec
measures.register(Rprec, ['RPrec'])



from ir_measures import measures
from .base import Measure, ParamInfo


class _RBP(measures.Measure):
    """
    The Rank-Biased Precision (RBP).


    .. code-block:: bibtex
        :caption: Citation

        @article{DBLP:journals/tois/MoffatZ08,
          author       = {Alistair Moffat and
                          Justin Zobel},
          title        = {Rank-biased precision for measurement of retrieval effectiveness},
          journal      = {{ACM} Trans. Inf. Syst.},
          volume       = {27},
          number       = {1},
          pages        = {2:1--2:27},
          year         = {2008},
          url          = {https://doi.org/10.1145/1416950.1416952},
          doi          = {10.1145/1416950.1416952}
        }
    """
    __name__ = 'RBP'
    NAME = __name__
    PRETTY_NAME = 'Rank-Biased Precision'
    SHORT_DESC = 'A measure of the rate at which utility is gained as a user traverses a result list with a degree of persistence.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'p': measures.ParamInfo(dtype=float, default=0.8, desc='persistence'),
        'rel': measures.ParamInfo(dtype=int, required=False, desc='minimum relevance score to be considered relevant (inclusive), or NOT_PROVIDED to use graded relevance')
    }


RBP = _RBP()
measures.register(RBP)



from ir_measures import measures
from .base import Measure, ParamInfo


class _R(measures.Measure):
    """
    Recall@k (R@k). The fraction of relevant documents for a query that have been retrieved by rank k.

    NOTE: Some tasks define Recall@k as whether any relevant documents are found in the top k results.
    This software follows the TREC convention and refers to that measure as Success@k.
    """
    __name__ = 'R'
    NAME = __name__
    PRETTY_NAME = 'Recall at k'
    SHORT_DESC = 'The percentage of relevant documents retrieved in the top k results.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


R = _R()
Recall = R
measures.register(R, ['Recall'])



from ir_measures import measures
from .base import Measure, ParamInfo, SumAgg


class _INST(measures.Measure):
    """
    INST, a variant of INSQ


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/sigir/BaileyMST15,
          author       = {Peter Bailey and
                          Alistair Moffat and
                          Falk Scholer and
                          Paul Thomas},
          title        = {User Variability and {IR} System Evaluation},
          booktitle    = {Proceedings of the 38th International {ACM} {SIGIR} Conference on
                          Research and Development in Information Retrieval, Santiago, Chile,
                          August 9-13, 2015},
          pages        = {625--634},
          publisher    = {{ACM}},
          year         = {2015},
          url          = {https://doi.org/10.1145/2766462.2767728},
          doi          = {10.1145/2766462.2767728}
        }
    """
    __name__ = 'INST'
    NAME = __name__
    PRETTY_NAME = 'INST'
    SHORT_DESC = 'An improved version of INSQ that better handles when either no documents or all retrieved documents are relevant.'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


class _INSQ(measures.Measure):
    """
    INSQ


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/adcs/MoffatST12,
          author       = {Alistair Moffat and
                          Falk Scholer and
                          Paul Thomas},
          title        = {Models and metrics: {IR} evaluation as a user process},
          booktitle    = {The Seventeenth Australasian Document Computing Symposium, {ADCS}
                          '12, Dunedin, New Zealand, December 5-6, 2012},
          pages        = {47--54},
          publisher    = {{ACM}},
          year         = {2012},
          url          = {https://doi.org/10.1145/2407085.2407092},
          doi          = {10.1145/2407085.2407092}
        }
    """
    __name__ = 'INSQ'
    NAME = __name__
    PRETTY_NAME = 'INSQ'
    SHORT_DESC = 'A weighted precision measure based on the conditional probability of the user continuing to the next item.'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


INST = _INST()
measures.register(INST)

INSQ = _INSQ()
measures.register(INSQ)



from ir_measures import measures
from .base import Measure, ParamInfo


class _NERR8(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (8) of the the following paper.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/ictir/AzzopardiMM21,
          author       = {Leif Azzopardi and
                          Joel Mackenzie and
                          Alistair Moffat},
          title        = {{ERR} is not {C/W/L:} Exploring the Relationship Between Expected
                          Reciprocal Rank and Other Metrics},
          booktitle    = {{ICTIR} '21: The 2021 {ACM} {SIGIR} International Conference on the
                          Theory of Information Retrieval, Virtual Event, Canada, July 11, 2021},
          pages        = {231--237},
          publisher    = {{ACM}},
          year         = {2021},
          url          = {https://doi.org/10.1145/3471158.3472239},
          doi          = {10.1145/3471158.3472239}
        }
    """
    __name__ = 'NERR8'
    NAME = __name__
    PRETTY_NAME = 'Nearly Expected Reciprocal Rank Eq 8'
    SHORT_DESC = 'A C/W/L approximation of ERR using gain-based stopping with truncation at k.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }

class _NERR9(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (9) of the the following paper.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/ictir/AzzopardiMM21,
          author       = {Leif Azzopardi and
                          Joel Mackenzie and
                          Alistair Moffat},
          title        = {{ERR} is not {C/W/L:} Exploring the Relationship Between Expected
                          Reciprocal Rank and Other Metrics},
          booktitle    = {{ICTIR} '21: The 2021 {ACM} {SIGIR} International Conference on the
                          Theory of Information Retrieval, Virtual Event, Canada, July 11, 2021},
          pages        = {231--237},
          publisher    = {{ACM}},
          year         = {2021},
          url          = {https://doi.org/10.1145/3471158.3472239},
          doi          = {10.1145/3471158.3472239}
        }
    """
    __name__ = 'NERR9'
    NAME = __name__
    PRETTY_NAME = 'Nearly Expected Reciprocal Rank Eq 9'
    SHORT_DESC = 'A C/W/L approximation of ERR using gain-based stopping and discount with truncation at k.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }

class _NERR10(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (10) of the the following paper.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/ictir/AzzopardiMM21,
          author       = {Leif Azzopardi and
                          Joel Mackenzie and
                          Alistair Moffat},
          title        = {{ERR} is not {C/W/L:} Exploring the Relationship Between Expected
                          Reciprocal Rank and Other Metrics},
          booktitle    = {{ICTIR} '21: The 2021 {ACM} {SIGIR} International Conference on the
                          Theory of Information Retrieval, Virtual Event, Canada, July 11, 2021},
          pages        = {231--237},
          publisher    = {{ACM}},
          year         = {2021},
          url          = {https://doi.org/10.1145/3471158.3472239},
          doi          = {10.1145/3471158.3472239}
        }
    """
    __name__ = 'NERR10'
    NAME = __name__
    PRETTY_NAME = 'Nearly Expected Reciprocal Rank Eq 10'
    SHORT_DESC = 'A C/W/L approximation of ERR using gain-based stopping and RBP patience (p).'
    SUPPORTED_PARAMS = {
        'p': measures.ParamInfo(dtype=float, default=0.9, desc='persistence'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }

class _NERR11(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (12) of the the following paper.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/ictir/AzzopardiMM21,
          author       = {Leif Azzopardi and
                          Joel Mackenzie and
                          Alistair Moffat},
          title        = {{ERR} is not {C/W/L:} Exploring the Relationship Between Expected
                          Reciprocal Rank and Other Metrics},
          booktitle    = {{ICTIR} '21: The 2021 {ACM} {SIGIR} International Conference on the
                          Theory of Information Retrieval, Virtual Event, Canada, July 11, 2021},
          pages        = {231--237},
          publisher    = {{ACM}},
          year         = {2021},
          url          = {https://doi.org/10.1145/3471158.3472239},
          doi          = {10.1145/3471158.3472239}
        }
    """
    __name__ = 'NERR11'
    NAME = __name__
    PRETTY_NAME = 'Nearly Expected Reciprocal Rank Eq 11'
    SHORT_DESC = 'A C/W/L approximation of ERR using gain based stopping and INST Goal (T).'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


NERR8 = _NERR8()
measures.register(NERR8)
NERR9 = _NERR9()
measures.register(NERR9)
NERR10 = _NERR10()
measures.register(NERR10)
NERR11 = _NERR11()
measures.register(NERR11)



from ir_measures import measures
from .base import BaseMeasure, ParamInfo


class _ERR_IA(measures.BaseMeasure):
    """
    Intent-Aware Expected Reciprocal Rank with collection-independent normalisation.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/cikm/ChapelleMZG09,
          author       = {Olivier Chapelle and
                          Donald Metlzer and
                          Ya Zhang and
                          Pierre Grinspan},
          title        = {Expected reciprocal rank for graded relevance},
          booktitle    = {Proceedings of the 18th {ACM} Conference on Information and Knowledge
                          Management, {CIKM} 2009, Hong Kong, China, November 2-6, 2009},
          pages        = {621--630},
          publisher    = {{ACM}},
          year         = {2009},
          url          = {https://doi.org/10.1145/1645953.1646033},
          doi          = {10.1145/1645953.1646033}
        }
    """
    __name__ = 'ERR_IA'
    NAME = __name__
    PRETTY_NAME = 'Intent-Aware Expected Reciprocal Rank'
    SHORT_DESC = 'A version of ERR that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _nERR_IA(measures.BaseMeasure):
    """
    Intent-Aware Expected Reciprocal Rank with collection-dependent normalisation.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/cikm/ChapelleMZG09,
          author       = {Olivier Chapelle and
                          Donald Metlzer and
                          Ya Zhang and
                          Pierre Grinspan},
          title        = {Expected reciprocal rank for graded relevance},
          booktitle    = {Proceedings of the 18th {ACM} Conference on Information and Knowledge
                          Management, {CIKM} 2009, Hong Kong, China, November 2-6, 2009},
          pages        = {621--630},
          publisher    = {{ACM}},
          year         = {2009},
          url          = {https://doi.org/10.1145/1645953.1646033},
          doi          = {10.1145/1645953.1646033}
        }
    """
    __name__ = 'nERR_IA'
    NAME = __name__
    PRETTY_NAME = 'Intent-Aware Normalised Expected Reciprocal Rank'
    SHORT_DESC = 'A normalised version of ERR that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _alpha_DCG(measures.BaseMeasure):
    """
    A version of DCG that accounts for multiple possible query intents.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/sigir/ClarkeKCVABM08,
          author       = {Charles L. A. Clarke and
                          Maheedhar Kolla and
                          Gordon V. Cormack and
                          Olga Vechtomova and
                          Azin Ashkan and
                          Stefan B{\"{u}}ttcher and
                          Ian MacKinnon},
          title        = {Novelty and diversity in information retrieval evaluation},
          booktitle    = {Proceedings of the 31st Annual International {ACM} {SIGIR} Conference
                          on Research and Development in Information Retrieval, {SIGIR} 2008,
                          Singapore, July 20-24, 2008},
          pages        = {659--666},
          publisher    = {{ACM}},
          year         = {2008},
          url          = {https://doi.org/10.1145/1390334.1390446},
          doi          = {10.1145/1390334.1390446}
        }
    """
    __name__ = 'alpha_DCG'
    NAME = __name__
    PRETTY_NAME = 'Alpha Discounted Cumulative Gain'
    SHORT_DESC = 'A version of DCG that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _alpha_nDCG(measures.BaseMeasure):
    """
    A version of nDCG that accounts for multiple possible query intents.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/sigir/ClarkeKCVABM08,
          author       = {Charles L. A. Clarke and
                          Maheedhar Kolla and
                          Gordon V. Cormack and
                          Olga Vechtomova and
                          Azin Ashkan and
                          Stefan B{\"{u}}ttcher and
                          Ian MacKinnon},
          title        = {Novelty and diversity in information retrieval evaluation},
          booktitle    = {Proceedings of the 31st Annual International {ACM} {SIGIR} Conference
                          on Research and Development in Information Retrieval, {SIGIR} 2008,
                          Singapore, July 20-24, 2008},
          pages        = {659--666},
          publisher    = {{ACM}},
          year         = {2008},
          url          = {https://doi.org/10.1145/1390334.1390446},
          doi          = {10.1145/1390334.1390446}
        }
    """
    __name__ = 'alpha_nDCG'
    NAME = __name__
    PRETTY_NAME = 'Alpha Normalised Discounted Cumulative Gain'
    SHORT_DESC = 'A version of nDCG that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _NRBP(measures.BaseMeasure):
    """
    Novelty- and Rank-Biased Precision with collection-independent normalisation.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/ictir/ClarkeKV09,
          author       = {Charles L. A. Clarke and
                          Maheedhar Kolla and
                          Olga Vechtomova},
          title        = {An Effectiveness Measure for Ambiguous and Underspecified Queries},
          booktitle    = {Advances in Information Retrieval Theory, Second International Conference
                          on the Theory of Information Retrieval, {ICTIR} 2009, Cambridge, UK,
                          September 10-12, 2009, Proceedings},
          series       = {Lecture Notes in Computer Science},
          volume       = {5766},
          pages        = {188--199},
          publisher    = {Springer},
          year         = {2009},
          url          = {https://doi.org/10.1007/978-3-642-04417-5\\_17},
          doi          = {10.1007/978-3-642-04417-5\\_17}
        }
    """
    __name__ = 'NRBP'
    NAME = __name__
    PRETTY_NAME = 'Novelty- and Rank-Biased Precision'
    SHORT_DESC = 'A version of RBP that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'beta': measures.ParamInfo(dtype=float, default=0.5, desc='Patience'),
    }

class _nNRBP(measures.BaseMeasure):
    """
    Novelty- and Rank-Biased Precision with collection-dependent normalisation.


    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/ictir/ClarkeKV09,
          author       = {Charles L. A. Clarke and
                          Maheedhar Kolla and
                          Olga Vechtomova},
          title        = {An Effectiveness Measure for Ambiguous and Underspecified Queries},
          booktitle    = {Advances in Information Retrieval Theory, Second International Conference
                          on the Theory of Information Retrieval, {ICTIR} 2009, Cambridge, UK,
                          September 10-12, 2009, Proceedings},
          series       = {Lecture Notes in Computer Science},
          volume       = {5766},
          pages        = {188--199},
          publisher    = {Springer},
          year         = {2009},
          url          = {https://doi.org/10.1007/978-3-642-04417-5\\_17},
          doi          = {10.1007/978-3-642-04417-5\\_17}
        }
    """
    __name__ = 'nNRBP'
    NAME = __name__
    PRETTY_NAME = 'Normalised Novelty- and Rank-Biased Precision'
    SHORT_DESC = 'A normalised version of RBP that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'beta': measures.ParamInfo(dtype=float, default=0.5, desc='Patience'),
    }

class _AP_IA(measures.BaseMeasure):
    """
    Intent-aware (Mean) Average Precision
    """
    __name__ = 'AP_IA'
    NAME = __name__
    PRETTY_NAME = 'Intent-Aware (Mean) Average Precision'
    SHORT_DESC = 'A version of AP that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _P_IA(measures.BaseMeasure):
    """
    Intent-aware Precision@k.
    """
    __name__ = 'P_IA'
    NAME = __name__
    PRETTY_NAME = 'Intent-Aware Precision@k'
    SHORT_DESC = 'A version of P@k that accounts for multiple possible query intents.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _StRecall(measures.BaseMeasure):
    """
    Subtopic recall (the number of subtopics covered by the top k docs)
    """
    __name__ = 'StRecall'
    NAME = __name__
    PRETTY_NAME = 'Subtopic Recall at k'
    SHORT_DESC = 'The percentage of subtopics covered by the top k documents.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
    }



ERR_IA = _ERR_IA()
nERR_IA = _nERR_IA()
alpha_DCG = _alpha_DCG()
α_DCG = alpha_DCG
alpha_nDCG = _alpha_nDCG()
α_nDCG = alpha_nDCG
NRBP = _NRBP()
nNRBP = _nNRBP()
AP_IA = _AP_IA()
MAP_IA = AP_IA
P_IA = _P_IA()
StRecall = _StRecall()

measures.register(ERR_IA)
measures.register(nERR_IA)
measures.register(alpha_DCG, aliases=['α_DCG'])
measures.register(alpha_nDCG, aliases=['α_nDCG'])
measures.register(NRBP)
measures.register(nNRBP)
measures.register(AP_IA, aliases=['MAP_IA'])
measures.register(P_IA)
measures.register(StRecall)



registry = {}
def register(measure, aliases=[], name=None):
    if name is None:
        name = measure.__name__
    assert name not in registry
    registry[name] = measure
    for alias in aliases:
        assert alias not in registry
        registry[alias] = measure
    return registry

from .accuracy import Accuracy, _Accuracy
from .base import Measure, BaseMeasure, ParamInfo, MultiMeasures, MeanAgg, SumAgg
from .ap import AP, MAP, _AP
from .bpm import BPM, _BPM
from .bpref import Bpref, BPref, _Bpref
from .compat import Compat, _Compat
from .diversity import ERR_IA, _ERR_IA, nERR_IA, _nERR_IA, alpha_DCG, α_DCG, _alpha_DCG, alpha_nDCG, α_nDCG, _alpha_nDCG, NRBP, _NRBP, nNRBP, _nNRBP, AP_IA, MAP_IA, _AP_IA, P_IA, _P_IA, StRecall, _StRecall
from .err import ERR, _ERR
from .inst import INST, _INST, INSQ, _INSQ
from .infap import infAP, _infAP
from .iprec import IPrec, _IPrec
from .judged import Judged, _Judged
from .ndcg import nDCG, NDCG, _nDCG
from .nerr import NERR8, NERR9, NERR10, NERR11, _NERR8, _NERR9, _NERR10, _NERR11
from .numq import NumQ, _NumQ
from .numrel import NumRel, _NumRel
from .numret import NumRet, NumRelRet, _NumRet
from .p import P, Precision, _P
from .r import R, Recall, _R
from .rbp import RBP, _RBP
from .rprec import Rprec, RPrec, _Rprec
from .rr import RR, MRR, _RR
from .sdcg import SDCG, _SDCG
from .set_measures import SetP, SetRelP, _SetP, SetR, _SetR, SetF, _SetF, SetAP, _SetAP
from .success import Success, _Success

# enable from "ir_measures.measures import *" --- on purpuse, do not include _-prefixed versions,
# as these are intended for internal use
__all__ = list(registry.keys())



from ir_measures import providers
from ir_measures.util import QrelsConverter, RunConverter, flatten_measures


class FallbackProvider(providers.Provider):
    def __init__(self, providers):
        super().__init__()
        self.providers = providers

    def _evaluator(self, measures, qrels):
        measures = flatten_measures(measures)
        orig_measures = list(measures)
        provider_measure_pairs = []
        provides_that_would_support = []
        for provider in self.providers:
            provider_measures = set()
            for measure in measures:
                if provider.supports(measure):
                    provider_measures.add(measure)
            if provider_measures:
                if not provider.is_available():
                    provides_that_would_support.append(provider)
                else:
                    measures = measures - provider_measures
                    provider_measure_pairs.append((provider, provider_measures))
                    if not measures:
                        break
        if measures:
            provider_message = ''
            if provides_that_would_support:
                if len(measures) == 1:
                    provider_message = ' The following providers would support this measure:'
                else:
                    provider_message = ' The following providers would support at least one of these measures:'
                for p in provides_that_would_support:
                    inst = p.install_instructions()
                    if inst:
                        provider_message += f'\n - {p.NAME} ({inst})'
                    else:
                        provider_message += f'\n - {p.NAME}'
            raise ValueError(f'Unsupported measures {measures}.{provider_message}')

        evaluators = []
        qrels_teed = QrelsConverter(qrels).tee(len(provider_measure_pairs))
        for (provider, provider_measures), qrels in zip(provider_measure_pairs, qrels_teed):
            evaluators.append(provider.evaluator(provider_measures, qrels.qrels))
        if len(evaluators) == 1:
            return evaluators[0] # skip the overhead of FallbackEvaluator if there's only one
        return FallbackEvaluator(orig_measures, evaluators)

    def supports(self, measure):
        return any(p.is_available() and p.supports(measure) for p in self.providers)


class FallbackEvaluator(providers.Evaluator):
    def __init__(self, measures, evaluators):
        super().__init__(measures, evaluators[0].qrel_qids)
        self.evaluators = evaluators

    def _iter_calc(self, run):
        runs_teed = RunConverter(run).tee(len(self.evaluators))
        for evaluator, run in zip(self.evaluators, runs_teed):
            yield from evaluator.iter_calc(run.run)



import contextlib
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED
from ir_measures.bin import msmarco_eval
import sys

class MsMarcoProvider(providers.Provider):
    """
    MS MARCO's implementation of RR
    """
    NAME = 'msmarco'
    SUPPORTED_MEASURES = [
        measures._RR(cutoff=Any(), rel=Any(), judged_only=Choices(False)),
    ]

    def _evaluator(self, measures, qrels):
        measures = ir_measures.util.flatten_measures(measures)

        invocations = []
        for measure in measures:
            if measure.NAME == 'RR':
                invocations.append((measure, measure['rel'], measure['cutoff']))
            else:
                raise ValueError(f'unsupported measure {measure}')

        return MsMarcoEvaluator(measures, qrels, invocations)


class MsMarcoEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invocations):
        query_ids = set()
        self.qrels_by_rel = {rel: {} for _, rel, _ in invocations}
        for qrel in ir_measures.util.QrelsConverter(qrels).as_namedtuple_iter():
            query_ids.add(qrel.query_id)
            for rel in self.qrels_by_rel:
                if qrel.relevance >= rel:
                    self.qrels_by_rel[rel].setdefault(qrel.query_id, {})[qrel.doc_id] = 1
        super().__init__(measures, query_ids)
        self.invocations = invocations

    def _iter_calc(self, run):
        run = ir_measures.util.RunConverter(run).as_dict_of_dict()
        sorted_run = {q: list(sorted(run[q].items(), key=lambda x: (-x[1], x[0]))) for q in run}
        sorted_run = {q: [did for did, _ in v] for q, v in sorted_run.items()}
        for measure, rel, cutoff in self.invocations:
            if cutoff is NOT_PROVIDED:
                cutoff = sys.maxsize
            msmarco_result = msmarco_eval.compute_metrics(self.qrels_by_rel[rel], sorted_run, max_rank=cutoff)
            for qid, value in msmarco_result[f'MRR @{cutoff} by query'].items():
                yield Metric(query_id=qid, measure=measure, value=value)


providers.register(MsMarcoProvider())



import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any


class CompatProvider(providers.Provider):
    """
    Version of the compatibility measure desribed in:

    .. code-block:: bibtex
        :caption: Citation

        @article{DBLP:journals/tois/ClarkeVS21,
          author       = {Charles L. A. Clarke and
                          Alexandra Vtyurina and
                          Mark D. Smucker},
          title        = {Assessing Top-k Preferences},
          journal      = {{ACM} Trans. Inf. Syst.},
          volume       = {39},
          number       = {3},
          pages        = {33:1--33:21},
          year         = {2021},
          url          = {https://doi.org/10.1145/3451161},
          doi          = {10.1145/3451161}
        }
    """
    NAME = 'compat'
    SUPPORTED_MEASURES = [
        measures._Compat(p=Any(), normalize=Any())
    ]

    def _evaluator(self, measures, qrels):
        invocations = []
        for measure in ir_measures.util.flatten_measures(measures):
            if measure.NAME == 'Compat':
                invocations.append((measure, measure['p'], measure['normalize']))
            else:
                raise ValueError(f'unsupported measure {measure}')
        qrels = ir_measures.util.QrelsConverter(qrels).as_dict_of_dict()
        return CompatEvaluator(measures, qrels, invocations)


def rbo(ranking, ideal, p, depth):
    ranking_set = set()
    ideal_set = set()
    score = 0.0
    normalizer = 0.0
    weight = 1.0
    for i in range(depth):
        if i < len(ranking):
            ranking_set.add(ranking[i])
        if i < len(ideal):
            ideal_set.add(ideal[i])
        score += weight*len(ideal_set.intersection(ranking_set))/(i + 1)
        normalizer += weight
        weight *= p
    return score/normalizer


def compatibility(qrels, run, p, normalize):
    ranking = list(run.keys())
    ranking.sort()
    ranking.sort(key=lambda docno: run[docno], reverse=True)
    ideal = [docno for docno in qrels if qrels[docno] > 0]
    ideal.sort(key=lambda docno: run[docno] if docno in run else 0.0, reverse=True)
    ideal.sort(key=lambda docno: qrels[docno], reverse=True)
    depth = max(len(ranking), len(ideal))
    score = rbo(ranking, ideal, p, depth)
    if normalize:
        best = rbo(ideal, ideal, p, depth)
        if best > 0.0:
            score = rbo(ranking, ideal, p, depth)/best
    return score


class CompatEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invocations):
        super().__init__(measures, set(qrels.keys()))
        self.qrels = qrels
        self.invocations = invocations

    def _iter_calc(self, run):
        run = ir_measures.util.RunConverter(run).as_dict_of_dict()
        for measure, p, normalize in self.invocations:
            for qid in run:
                if qid in self.qrels:
                    value = compatibility(self.qrels[qid], run[qid], p, normalize)
                    yield Metric(query_id=qid, measure=measure, value=value)


providers.register(CompatProvider())



import unittest
import itertools
import ir_measures


class TestMeasures(unittest.TestCase):

    def test_P(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 2
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
1 0 D0 5 0.1 run
'''))
        measure = ir_measures.P@5
        result = measure.calc_aggregate(qrels, run)
        self.assertEqual(result, 0.5)
        agg_result, perq_result = measure.calc(qrels, run)
        self.assertEqual(agg_result, 0.5)
        self.assertEqual(len(perq_result), 2)


if __name__ == '__main__':
    unittest.main()



from ir_measures import measures
from .base import Measure, ParamInfo


class _Compat(measures.Measure):
    """
    Compatibility measure desribed in:


    .. code-block:: bibtex
        :caption: Citation

        @article{DBLP:journals/tois/ClarkeVS21,
          author       = {Charles L. A. Clarke and
                          Alexandra Vtyurina and
                          Mark D. Smucker},
          title        = {Assessing Top-k Preferences},
          journal      = {{ACM} Trans. Inf. Syst.},
          volume       = {39},
          number       = {3},
          pages        = {33:1--33:21},
          year         = {2021},
          url          = {https://doi.org/10.1145/3451161},
          doi          = {10.1145/3451161}
        }
    """
    __name__ = 'Compat'
    NAME = __name__
    PRETTY_NAME = 'Compatibility'
    SHORT_DESC = 'The Rank Biased Overlap between the results and an ideal ranking.'
    SUPPORTED_PARAMS = {
        'p': measures.ParamInfo(dtype=float, default=0.95, desc='persistence'),
        'normalize': measures.ParamInfo(dtype=bool, default=True, desc='apply normalization for finite ideal rankings'),
    }


Compat = _Compat()
measures.register(Compat)



from ir_measures import measures
from .base import Measure, ParamInfo, SumAgg


class _NumRel(measures.Measure):
    """
    The number of relevant documents the query has (independent of what the system retrieved).
    """
    __name__ = 'NumRel'
    NAME = __name__
    PRETTY_NAME = 'Number of Relevant Documents'
    SHORT_DESC = 'The number of relevant documents present in the qrels'
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be counted (inclusive)')
    }

    def aggregator(self):
        return SumAgg()


NumRel = _NumRel()
measures.register(NumRel)



import unittest
import itertools
import ir_measures
from ir_measures import Metric


class TestRanx(unittest.TestCase):

    def test_P(self):
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
'''))
        measure = ir_measures.P@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6))
        measure = ir_measures.P(rel=2)@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.2))

        measure = ir_measures.SetP
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6))
        measure = ir_measures.SetP(rel=2)@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.2))

        measure = ir_measures.R@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.0))
        measure = ir_measures.R(rel=2)@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.0))
        measure = ir_measures.R@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.3333333333333333))
        measure = ir_measures.R(rel=2)@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.0))

        measure = ir_measures.SetR
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.0))
        measure = ir_measures.SetR(rel=2)@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.0))

        measure = ir_measures.RR
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.5))
        measure = ir_measures.RR(rel=2)
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.3333333333333333))
        measure = ir_measures.RR@10
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.5))
        measure = ir_measures.RR(rel=2)@10
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.3333333333333333))
        measure = ir_measures.RR@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.5))
        measure = ir_measures.RR(rel=2)@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.))

        measure = ir_measures.AP
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6388888888888888))
        measure = ir_measures.AP(rel=2)
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.3333333333333333))
        measure = ir_measures.AP@10
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6388888888888888))
        measure = ir_measures.AP(rel=2)@10
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.3333333333333333))
        measure = ir_measures.AP@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.16666666666666666))
        measure = ir_measures.AP(rel=2)@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.0))

        measure = ir_measures.Success@10
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.))
        measure = ir_measures.Success(rel=2)@10
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.))
        measure = ir_measures.Success@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 1.))
        measure = ir_measures.Success(rel=2)@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.))

        measure = ir_measures.NumRet(rel=1)
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 3.))
        measure = ir_measures.NumRet(rel=2)
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))

        measure = ir_measures.nDCG
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6584645692843067))
        measure = ir_measures.nDCG@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6584645692843067))
        measure = ir_measures.nDCG@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.23981246656813146))

        measure = ir_measures.nDCG(dcg='exp-log2')
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6201040599710453))
        measure = ir_measures.nDCG(dcg='exp-log2')@5
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6201040599710453))
        measure = ir_measures.nDCG(dcg='exp-log2')@2
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.17376534287144002))

        measure = ir_measures.Rprec
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.6666666666666666))
        measure = ir_measures.Rprec(rel=2)
        result = list(ir_measures.ranx.iter_calc([measure], qrels, run))
        self.assertEqual(result[0], Metric('0', measure, 0.))


if __name__ == '__main__':
    unittest.main()



from ir_measures import measures
from .base import Measure, ParamInfo


class _RR(measures.Measure):
    """
    The [Mean] Reciprocal Rank ([M]RR) is a precision-focused measure that scores based on the reciprocal of the rank of the
    highest-scoring relevance document. An optional cutoff can be provided to limit the
    depth explored. rel (default 1) controls which relevance level is considered relevant.


    .. code-block:: bibtex
        :caption: Citation

        @article{DBLP:journals/ir/KantorV00,
          author       = {Paul B. Kantor and
                          Ellen M. Voorhees},
          title        = {The {TREC-5} Confusion Track: Comparing Retrieval Methods for Scanned
                          Text},
          journal      = {Inf. Retr.},
          volume       = {2},
          number       = {2/3},
          pages        = {165--176},
          year         = {2000},
          url          = {https://doi.org/10.1023/A:1009902609570},
          doi          = {10.1023/A:1009902609570}
        }
    """
    __name__ = 'RR'
    NAME = __name__
    PRETTY_NAME = '(Mean) Reciprocal Rank'
    SHORT_DESC = 'The reciprocal of the rank of the first relevant document.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


RR = _RR()
MRR = RR
measures.register(RR, ['MRR'])



from ir_measures import measures
from .base import Measure, ParamInfo


class _BPM(measures.Measure):
    """
    The Bejeweled Player Model (BPM).

    .. code-block:: bibtex
        :caption: Citation

        @inproceedings{DBLP:conf/sigir/ZhangLLZXM17,
          author       = {Fan Zhang and
                          Yiqun Liu and
                          Xin Li and
                          Min Zhang and
                          Yinghui Xu and
                          Shaoping Ma},
          title        = {Evaluating Web Search with a Bejeweled Player Model},
          booktitle    = {Proceedings of the 40th International {ACM} {SIGIR} Conference on
                          Research and Development in Information Retrieval, Shinjuku, Tokyo,
                          Japan, August 7-11, 2017},
          pages        = {425--434},
          publisher    = {{ACM}},
          year         = {2017},
          url          = {https://doi.org/10.1145/3077136.3080841},
          doi          = {10.1145/3077136.3080841}
        }
    """
    __name__ = 'BPM'
    NAME = __name__
    PRETTY_NAME = 'Bejeweled Player Model'
    SHORT_DESC = 'A measure that balances both gain and user patience to determine when they stop traversing search results.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'T': measures.ParamInfo(dtype=float, default=1., desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


BPM = _BPM()
measures.register(BPM)



import warnings
import contextlib
import itertools
from typing import Iterator, Iterable, Dict, Union
from ir_measures import measures, Metric, Qrel, ScoredDoc, CalcResults, Measure


class Evaluator:
    """
    The base class for scoring runs for a given set of measures and qrels.
    Returned from :meth:`Provider.evaluator() <ir_measures.providers.Provider.evaluator>`.
    """
    def __init__(self, measures: Iterable[Measure], qrel_qids: Iterable[str]):
        self.measures = measures
        self.qrel_qids = qrel_qids

    def iter_calc(self, run: Iterable[ScoredDoc]) -> Iterator[Metric]:
        """
        Yields per-topic metrics this run.
        """
        expected_measure_qids = set(itertools.product(self.measures, self.qrel_qids))
        for metric in self._iter_calc(run):
            expected_measure_qids.discard((metric.measure, metric.query_id))
            yield metric
        for measure, query_id in sorted(expected_measure_qids, key=lambda x: (str(x[0]), x[1])):
            yield Metric(query_id=query_id, measure=measure, value=measure.DEFAULT)

    def _iter_calc(self, run: Iterable[ScoredDoc]) -> Iterator[Metric]:
        raise NotImplementedError()

    def calc_aggregate(self, run: Iterable[ScoredDoc]) -> Dict[measures.Measure, Union[float, int]]:
        """
        Returns aggregated measure values for this run.
        """
        aggregators = {m: m.aggregator() for m in self.measures}
        for metric in self.iter_calc(run):
            aggregators[metric.measure].add(metric.value)
        return {m: agg.result() for m, agg in aggregators.items()}

    def calc(self, run: Iterable[ScoredDoc]) -> CalcResults:
        """
        Returns aggregated and per-query results for this run.
        """
        aggregators = {m: m.aggregator() for m in self.measures}
        metrics = []
        for metric in self.iter_calc(run):
            aggregators[metric.measure].add(metric.value)
            metrics.append(metric)
        agg = {m: agg.result() for m, agg in aggregators.items()}
        return CalcResults(agg, metrics)


class Provider:
    """
    The base class for all measure providers (e.g., :ref:`providers.pytrec_eval`, :ref:`providers.gdeval`, etc.).

    A ``Provider`` implements the calculation logic for one or more :class:`~ir_measures.measures.Measure` (e.g.,
    :ref:`measures.nDCG`, :ref:`measures.P`, etc.).
    """
    NAME = None
    SUPPORTED_MEASURES = []

    def __init__(self):
        self._is_available = None

    def _check_available(self):
        if not self.is_available():
            instr = self.install_instructions()
            if instr:
                instr = f' To install:\n  {instr}'
            raise RuntimeError(f'Provider {self.NAME} is not available.{instr}')

    def evaluator(self, measures: Iterable[Measure], qrels: Iterable[Qrel]) -> Evaluator:
        """
        Returns an :class:`~ir_measures.providers.Evaluator` for these measures and qrels, which
        can efficiently process multiple runs.
        """
        self._check_available()
        return self._evaluator(measures, qrels)

    @contextlib.contextmanager
    def calc_ctxt(self, measures, qrels):
        warnings.warn("calc_ctxt deprecated in 0.2.0. Please use .evaluator(measures, qrels) instead.", DeprecationWarning)
        self._check_available()
        evaluator = self._evaluator(measures, qrels)
        def _eval(run):
            yield from evaluator.iter_calc(run)
        yield _eval

    def iter_calc(self, measures: Iterable[Measure], qrels, run: Iterable[ScoredDoc]) -> Iterator[Metric]:
        """
        Yields per-topic metrics for these measures, qrels, and run.
        """
        self._check_available()
        return self._iter_calc(measures, qrels, run)

    def _evaluator(self, measures: Iterable[Measure], qrels: Iterable[Qrel]):
        raise NotImplementedError()

    def _iter_calc(self, measures: Iterable[Measure], qrels: Iterable[Qrel], run: Iterable[ScoredDoc]):
        return self._evaluator(measures, qrels).iter_calc(run)

    def calc_aggregate(self, measures: Iterable[Measure], qrels: Iterable[Qrel], run: Iterable[ScoredDoc]) -> Dict[Measure, Union[float, int]]:
        """
        Returns aggregated measure values for these measures, qrels, and run.
        """
        return self.evaluator(measures, qrels).calc_aggregate(run)

    def calc(self, measures: Iterable[Measure], qrels: Iterable[Qrel], run: Iterable[ScoredDoc]) -> CalcResults:
        """
        Returns aggregated and per-query results for these measures, qrels, and run.
        """
        return self.evaluator(measures, qrels).calc(run)

    def supports(self, measure):
        measure.validate_params()
        for supported_measure in self.SUPPORTED_MEASURES:
            if supported_measure.NAME == measure.NAME:
                valid = True
                for param_name, param_spec in supported_measure.params.items():
                    if not param_spec.validate(measure[param_name]):
                        valid = False
                        break
                if valid:
                    return True
        return False

    def is_available(self):
        if self._is_available is not None:
            return self._is_available
        try:
            self.initialize()
            self._is_available = True
        except RuntimeError:
            self._is_available = False
        return self._is_available

    def initialize(self):
        pass

    def install_instructions(self):
        return None


class ParamSpec:
    def validate(self, value):
        raise NotImplementedError()

class Any:
    def __init__(self, required=False):
        self.required = required

    def validate(self, value):
        if self.required:
            return value is not NOT_PROVIDED
        return True

    def __repr__(self):
        if not self.required:
            return 'ANY'
        return 'REQUIRED'

class Choices:
    def __init__(self, *args):
        self.choices = args

    def validate(self, value):
        return value in self.choices

    def __repr__(self):
        if len(self.choices) == 1:
            if self.choices[0] == NOT_PROVIDED:
                return 'NOT_PROVIDED'
            return repr(self.choices[0])
        return repr(self.choices)

NOT_PROVIDED = measures.base._NOT_PROVIDED



from ir_measures import measures
from .base import Measure, ParamInfo

class _Accuracy(Measure):
    """
    Reports the probability that a relevant document is ranked before a non relevant one.
    This metric purpose is to be used for diagnosis (checking that train/test/validation accuracy match).
    As such, it only considers relevant documents which are within the returned ones.
    """
    __name__ = 'Accuracy'
    NAME = __name__
    PRETTY_NAME = 'Accuracy'
    SHORT_DESC = 'The probability that a relevant document is ranked before a non relevant one.'
    SUPPORTED_PARAMS = {
        'cutoff': ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }

Accuracy = _Accuracy()
measures.register(Accuracy)



import itertools
import ir_measures
from ir_measures import CalcResults


class Measure:
    NAME = None
    AT_PARAM = 'cutoff' # allows measures to configure which param measure@X updates (default is cutoff)
    SUPPORTED_PARAMS = {}
    DEFAULT = 0. # value if no documents are returned for this query

    def __init__(self, **params):
        self.params = params
        self.validated = False

    def validate_params(self):
        if self.validated:
            return
        assert isinstance(self.params, dict), "params must be dict"
        unsupposed_params = list(self.params.keys() - self.SUPPORTED_PARAMS.keys())
        assert len(unsupposed_params) == 0, f"unsupported params found: {unsupposed_params}"
        for param_name, param in self.SUPPORTED_PARAMS.items():
            param_val = self.params.get(param_name, _NOT_PROVIDED)
            is_valid = param.validate(param_val)
            assert is_valid, f"invalid param {param_name}={repr(param_val)}"
        self.validated = True

    def __call__(self, **kwargs):
        product = []
        for k, v in kwargs.items():
            if isinstance(v, (list, tuple)):
                product.append([(k, item) for item in v])
        results = []
        for items in itertools.product(*product):
            params = dict(self.params)
            params.update(kwargs)
            params.update(items)
            results.append(type(self)(**params))
        if product:
            result = MultiMeasures(*results)
            return result
        return results[0]

    def __matmul__(self, at_param):
        return self(**{self.AT_PARAM: at_param})

    def __getitem__(self, key):
        default = self.SUPPORTED_PARAMS[key].default
        return self.params.get(key, default)

    def iter_calc(self, qrels, run):
        self.validate_params()
        return ir_measures.iter_calc([self], qrels, run)

    def calc_aggregate(self, qrels, run):
        return ir_measures.calc_aggregate([self], qrels, run)[self]

    def calc(self, qrels, run) -> CalcResults:
        agg, perq = ir_measures.calc([self], qrels, run)
        return CalcResults(agg[self], perq)

    def evaluator(self, qrels):
        return ir_measures.evaluator([self], qrels)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        result = self.__name__
        params = ','.join(f'{k}={self._param_repr(v)}' for k, v in self.params.items() if k != self.AT_PARAM and v != self.SUPPORTED_PARAMS[k].default)
        if params:
            result = f'{result}({params})'
        if self.AT_PARAM in self.params:
            result = f'{result}@{self.params[self.AT_PARAM]}'
        return result

    def _param_repr(self, v):
        if isinstance(v, dict):
            return '{' + ','.join(f'{k}:{v}' for k, v in sorted(v.items()) if k != v) + '}'
        return repr(v)

    def __eq__(self, other):
        if isinstance(other, Measure):
            return repr(self) == repr(other)
        return False

    def __hash__(self):
        return hash(repr(self))

    def aggregator(self):
        return MeanAgg()


BaseMeasure = Measure # for compatibility


class MeanAgg:
    def __init__(self, default=float('NaN')):
        self.sum = 0.
        self.count = 0
        self.default = default

    def add(self, value):
        self.sum += value
        self.count += 1

    def result(self):
        if self.count == 0:
            return self.default
        return self.sum / self.count


class SumAgg:
    def __init__(self):
        self.sum = 0

    def add(self, value):
        self.sum += value

    def result(self):
        return self.sum


class MultiMeasures:
    def __init__(self, *measures):
        self.measures = set()
        self._add_measures(measures)

    def _add_measures(self, measures):
        for m in measures:
            if isinstance(m, MultiMeasures):
                self._add_measures(m.measures)
            else:
                self.measures.add(m)

    def __call__(self, **kwargs):
        return MultiMeasures(*(m(**kwargs) for m in self.measures))

    def __matmul__(self, at_param):
        return MultiMeasures(*(m(**{m.AT_PARAM: at_param}) for m in self.measures))

    def __str__(self):
        return repr(self)

    def __repr__(self):
        if self.measures:
            return f'MultiMeasures({repr(self.measures)[1:-1]})'
        return 'MultiMeasures()'

    def iter_calc(self, qrels, run):
        return ir_measures.DefaultPipeline.iter_calc(self.measures, qrels, run)

    def calc_aggregate(self, qrels, run):
        return self.aggregate(self.iter_calc())


_NOT_PROVIDED = object()


class ParamInfo:
    def __init__(self, dtype=None, required=False, choices=_NOT_PROVIDED, default=_NOT_PROVIDED, desc=None):
        self.dtype = dtype
        self.required = required
        self.choices = choices
        self.default = default
        self.desc = desc

    def validate(self, value):
        if value is _NOT_PROVIDED:
            return not self.required
        if self.dtype is not None and not isinstance(value, self.dtype):
            return False
        if self.choices is not _NOT_PROVIDED and value not in self.choices:
            return False
        return True
